import hmac
import hashlib

from django_pay2.settings import payment_settings
from django_pay2.models import Payment
from django_pay2.utils import MappingFieldsMixin

from rest_framework import serializers

from .constants import SberbankCallbackOperation, SberbankCallbackStatus


class SberbankCallbackSerializer(MappingFieldsMixin, serializers.Serializer):
    # TODO: посмотреть на amount и почему он integer
    amount = serializers.IntegerField()
    md_order = serializers.CharField()
    order_num = serializers.PrimaryKeyRelatedField(queryset=Payment.objects.all())
    checksum = serializers.CharField()
    operation = serializers.ChoiceField(SberbankCallbackOperation.choices)
    status = serializers.ChoiceField(SberbankCallbackStatus.choices)

    default_error_messages = {
        "invalid_checksum": 'Expected "{expected}", given "{given}"',
        "mismatch_amount": "Expected {expected}, given {given}",
    }

    mapping_keys = {
        "md_order": "mdOrder",
        "order_num": "orderNumber",
    }

    def validate(self, attrs):
        attrs = super().validate(attrs)
        payment = attrs["order_num"]
        if attrs["amount"] != payment.amount:
            self.fail("mismatch_amount", expected=payment.amount, given=attrs["amount"])
        return attrs

    def validate_checksum(self, given):
        checksum_data = self.initial_data.copy()
        checksum_data.pop("checksum")
        checksum_data.pop("sign_alias", None)
        checksum_str = (
            ";".join(
                f"{key};{checksum_data[key]}" for key in sorted(checksum_data.keys())
            )
            + ";"
        )
        expected = (
            hmac.new(
                payment_settings.SBERBANK.public_key.encode("utf-8"),
                checksum_str.encode("utf-8"),
                hashlib.sha256,
            )
            .hexdigest()
            .upper()
        )
        if given != expected:
            self.fail("invalid_checksum", expected=expected, given=given)
        return given



from dataclasses import dataclass
from typing import Dict


@dataclass
class PaymentRedirect:
    method = "redirect"
    url: str


@dataclass
class PaymentForm:
    method = "form"
    action: str
    fields: Dict[str, str]



import requests
import hmac
import hashlib

from django_pay2.payment_methods import PaymentRedirect
from django_pay2.settings import payment_settings


class QiwiKassaApi:
    def __init__(self, secret_key, site_id):
        self.secret_key = secret_key
        self.site_id = site_id

    def generate_payment_method(self, amount, currency, expirate_at, bill_id):
        response = requests.put(
            f"https://api.qiwi.com/partner/payin/v1/sites/{self.site_id}/bills/{bill_id}",
            json={
                "amount": {
                    "value": f"{amount:.2f}",
                    "currency": currency,
                },
                "expirationDateTime": expirate_at.strftime("%Y-%m-%dT%H:%M:%S+03:00"),
            },
            headers={"Authorization": f"Bearer {self.secret_key}"},
        )
        response.raise_for_status()
        return PaymentRedirect(response.json()["payUrl"])

    def create_hmac(self, payment_id, created_datetime, amount):
        return hmac.new(
            self.secret_key.encode("utf-8"),
            f"{payment_id}|{created_datetime}|{amount}",
            hashlib.sha256,
        ).hexdigest()


def get_api():
    return QiwiKassaApi(
        payment_settings.QIWI_KASSA.secret_key,
        payment_settings.QIWI_KASSA.site_id,
    )



from rest_framework import serializers

from django_pay2.models import Payment


class CoinPaymentsApproveSerializer(serializers.Serializer):
    invoice = serializers.PrimaryKeyRelatedField(queryset=Payment.objects.all())
    status = serializers.IntegerField()

    default_error_messages = {"not_done_status": "Not done status"}

    def validate_status(self, value):
        if not (value >= 100 or value == 2):
            self.fail("not_done_status")
        return value



from rest_framework import serializers


class PaymentRedirectSerializer(serializers.Serializer):
    method = serializers.CharField()
    url = serializers.CharField()


class PaymentFormSerializer(serializers.Serializer):
    method = serializers.CharField()
    action = serializers.CharField()
    fields = serializers.DictField()



# Generated by Django 3.1.6 on 2021-02-17 06:15

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=20, verbose_name='Сумма')),
                ('status', models.PositiveSmallIntegerField(choices=[(0, 'Ожидает оплаты'), (1, 'Успешно'), (2, 'Отклонен')], default=0, verbose_name='Статус')),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Дата создания')),
                ('object_id', models.UUIDField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
            ],
            options={
                'verbose_name': 'Платеж',
                'verbose_name_plural': 'Платежи',
                'ordering': ['-created'],
            },
        ),
    ]



from .exceptions import TinkoffNetworkError, TinkoffResponseError
from dataclasses import dataclass
from django_pay2.utils import clear_none
from typing import Optional, SupportsInt, Iterable
import hashlib
import requests
from .constants import (
    Languages,
    PayTypes,
    Taxations,
    Taxes,
    PaymentObjects,
    PaymentMethods,
)


def _bool_to_str(a_bool):
    return str(a_bool).lower()


def _to_json_str(value):
    if isinstance(value, bool):
        return _bool_to_str(value)
    return str(value)


class TinkoffApi:
    def __init__(self, terminal_key: str, password: str):
        self.terminal_key = terminal_key
        self.password = password

    def init_payment(
        self,
        amount_rub: SupportsInt,
        order_id: str,
        ip: Optional[str] = None,
        description: Optional[str] = None,
        lang: Optional[Languages] = None,
        notification_url: Optional[str] = None,
        success_url: Optional[str] = None,
        fail_url: Optional[str] = None,
        pay_type: Optional[PayTypes] = None,
        items: Optional[Iterable["TinkoffOrderItem"]] = None,
        client_email: Optional[str] = None,
        client_phone: Optional[str] = None,
        email_company: Optional[str] = None,
        taxation: Optional[Taxations] = None,
    ):
        token_data = clear_none(
            {
                "TerminalKey": self.terminal_key,
                "Amount": int(amount_rub * 100),
                "OrderId": order_id,
                "IP": ip,
                "Description": description,
                "Language": lang,
                "NotificationURL": notification_url,
                "SuccessURL": success_url,
                "FailURL": fail_url,
                "PayType": pay_type,
                "Receipt": clear_none(
                    {
                        "Items": [
                            clear_none(
                                {
                                    "Name": item.name,
                                    "Quantity": item.quantity,
                                    "Amount": int(item.amount_rub * 100),
                                    "Price": int(item.price_rub * 100),
                                    "PaymentMethod": item.payment_method,
                                    "PaymentObject": item.payment_object,
                                    "Tax": item.tax,
                                }
                            )
                            for item in items
                        ],
                        "Email": client_email,
                        "Phone": client_phone,
                        "EmailCompany": email_company,
                        "Taxation": taxation,
                    }
                )
                if items is not None
                else None,
            }
        )
        data = {
            **token_data,
            "Token": self.calculate_token(token_data),
        }
        raw_response = requests.post("https://securepay.tinkoff.ru/v2/Init", json=data)
        if not raw_response.ok:
            raise TinkoffNetworkError(raw_response.status_code, raw_response.text)

        response = InitResponse(raw_response.json())
        if not response.is_success:
            raise TinkoffResponseError(
                response.error_code, response.message, response.details
            )
        return response

    def calculate_token(self, data):
        prep_data = {
            k: v for k, v in data.items() if k not in ["Receipt", "Data", "Token"]
        }
        prep_data["Password"] = self.password
        sorted_data = {k: prep_data[k] for k in sorted(prep_data.keys())}
        hash_str = "".join(_to_json_str(v) for v in sorted_data.values())
        hash_obj = hashlib.sha256(hash_str.encode("utf-8"))
        return hash_obj.hexdigest()


class InitResponse:
    def __init__(self, raw_data):
        self.is_success: str = raw_data["Success"]
        self.payment_id: int = raw_data.get("PaymentId")
        self.error_code: str = raw_data["ErrorCode"]
        self.payment_url: Optional[str] = raw_data.get("PaymentURL")
        self.message: Optional[str] = raw_data.get("Message")
        self.details: Optional[str] = raw_data.get("Details")


@dataclass
class TinkoffOrderItem:
    name: str
    quantity: int
    amount_rub: SupportsInt
    price_rub: SupportsInt
    tax: Taxes
    payment_method: Optional[PaymentMethods] = None
    payment_object: Optional[PaymentObjects] = None

    def serialize(self):
        return clear_none(
            {
                "Name": self.name,
                "Quantity": self.quantity,
                "Amount": int(self.amount_rub * 100),
                "Price": int(self.price_rub * 100),
                "PaymentMethod": self.payment_method,
                "PaymentObject": self.payment_object,
                "Tax": self.tax,
            }
        )



from django.http import Http404
from django.shortcuts import redirect
from django.views import generic
from django.views.generic.detail import SingleObjectMixin

from .models import Payment
from .settings import payment_settings


class DebugMixin(generic.View):
    def dispatch(self, request, *args, **kwargs):
        if not payment_settings.DEBUG_MODE:
            raise Http404
        return super().dispatch(request, *args, **kwargs)


class DebugPaymentView(DebugMixin, generic.DetailView):
    queryset = Payment.objects.filter(status=Payment.StatusType.PENDING)
    template_name = "django_pay2/debug_payment.html"

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)


class AcceptDebugPaymentView(DebugMixin, SingleObjectMixin, generic.View):
    queryset = Payment.objects.filter(status=Payment.StatusType.PENDING)

    def post(self, request, *args, **kwargs):
        payment = self.get_object()
        payment.accept()
        return redirect("django_pay2:success")


class RejectDebugPaymentView(DebugMixin, SingleObjectMixin, generic.View):
    queryset = Payment.objects.filter(status=Payment.StatusType.PENDING)

    def post(self, request, *args, **kwargs):
        payment = self.get_object()
        payment.reject()
        return redirect("django_pay2:fail")


class SuccessPaymentView(generic.TemplateView):
    def get_template_names(self):
        return [payment_settings.TEMPLATES.success]


class RejectedPaymentView(generic.TemplateView):
    def get_template_names(self):
        return [payment_settings.TEMPLATES.rejected]



import logging

from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.views import generic
from django.views.decorators.csrf import csrf_exempt

from .functions import get_payeer_api
from .exceptions import PayeerValidationError, AlreadyPaid


logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name="dispatch")
class NotifyView(generic.View):
    def post(self, request, *args, **kwargs):
        api = get_payeer_api()
        try:
            result = api.notify(request.POST)
        except PayeerValidationError as exc:
            logger.info(f'Payeer validation error for payment {exc.order_id}: {exc.msg}')
            return HttpResponse(f"{exc.order_id}|error")
        except AlreadyPaid as exc:
            return HttpResponse(f"{exc.order_id}|success")

        result.payment.accept()
        logger.info(f'Successfully payeer paid {result.payment.id}')
        return HttpResponse(f"{result.raw_order_id}|success")



from rest_framework import serializers

from django_pay2.models import Payment
from django_pay2.utils import HmacDefault

from .api import get_api


class StatusSerializer(serializers.Serializer):
    value = serializers.CharField()

    default_error_messages = {"unexpected_status": "Unexpected status"}

    def validate_value(self, value):
        if value != "PAID":
            self.fail("unexpected_status")
        return value


class AmountSerializer(serializers.Serializer):
    value = serializers.DecimalField(max_digits=20, decimal_places=2)
    currency = serializers.CharField()


class BillSerializer(serializers.Serializer):
    site_id = serializers.CharField()
    bill_id = serializers.PrimaryKeyRelatedField(queryset=Payment.objects.all())
    amount = AmountSerializer()
    status = StatusSerializer()

    default_error_messages = {"mismatched_amount": "Mismatched amount"}

    def validate(self, attrs):
        attrs = super().validate(attrs)
        bill = attrs["bill_id"]
        if bill.amount != attrs["amount"]["value"]:
            self.fail("mismatched_amount")
        return attrs

    def to_internal_value(self, data):
        data = {
            "bill_id": data.get("billId"),
            "amount": data.get("amount"),
            "status": data.get("status"),
            "site_id": data.get("siteId"),
        }
        return super().to_internal_value(data)


class QiwiNotifySerializer(serializers.Serializer):
    bill = BillSerializer()
    hmac = serializers.HiddenField(default=HmacDefault())

    default_error_messages = {"mismatch_hmac": "Mismatch HMAC"}

    def validate(self, attrs):
        attrs = super().validate(attrs)
        expected_hmac = get_api().create_hmac(
            attrs["bill"]["amount"]["currency"],
            str(attrs["bill"]["amount"]["value"]),
            str(attrs["bill"]["bill_id"].id),
            attrs["bill"]["site_id"],
            attrs["bill"]["status"]["value"],
        )
        if expected_hmac != attrs["hmac"]:
            self.fail("mismatch_hmac")
        return attrs
